# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import numpy as np
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.meteo import constants
from earthkit.meteo import thermo


# Alternative proposed by Baudouin Raoult
class AutoDict(dict):
    def __missing__(self, key):
        value = self[key] = type(self)()
        return value


class NewDataField:
    def __init__(self, field, data, new_name):
        self.field = field
        self.data = data
        self.new_name = new_name

    def to_numpy(self, *args, **kwargs):
        return self.data

    def metadata(self, key=None, **kwargs):
        if key is None:
            return self.field.metadata(**kwargs)

        value = self.field.metadata(key, **kwargs)
        if key == "param":
            return self.new_name
        return value

    def __getattr__(self, name):
        return getattr(self.field, name)


def model_level_pressure(A, B, surface_pressure):
    """Calculates:
     - pressure at the model full- and half-levels
     - delta: depth of log(pressure) at full levels
     - alpha: alpha term #TODO: more descriptive information

    Parameters
    ----------
    A : ndarray
        A-coefficients defining the model levels
    B : ndarray
        B-coefficients defining the model levels
    surface_pressure: number or ndarray
        surface pressure (Pa)

    Returns
    -------
    ndarray
        pressure at model full-levels
    ndarray
        pressure at model half-levels
    ndarray
        delta at full-levels
    ndarray
        alpha at full levels
    """

    # constants
    PRESSURE_TOA = 0.1  # safety when highest pressure level = 0.0

    # make the calculation agnostic to the number of dimensions
    ndim = surface_pressure.ndim
    new_shape_half = (A.shape[0],) + (1,) * ndim
    A_reshaped = A.reshape(new_shape_half)
    B_reshaped = B.reshape(new_shape_half)

    # calculate pressure on model half-levels
    p_half_level = A_reshaped + B_reshaped * surface_pressure[np.newaxis, ...]

    # calculate delta
    new_shape_full = (A.shape[0] - 1,) + surface_pressure.shape
    delta = np.zeros(new_shape_full)
    delta[1:, ...] = np.log(p_half_level[2:, ...] / p_half_level[1:-1, ...])

    # pressure at highest half level<= 0.1
    if np.any(p_half_level[0, ...] <= PRESSURE_TOA):
        delta[0, ...] = np.log(p_half_level[1, ...] / PRESSURE_TOA)
    # pressure at highest half level > 0.1
    else:
        delta[0, ...] = np.log(p_half_level[1, ...] / p_half_level[0, ...])

    # calculate alpha
    alpha = np.zeros(new_shape_full)

    alpha[1:, ...] = 1.0 - p_half_level[1:-1, ...] / (p_half_level[2:, ...] - p_half_level[1:-1, ...]) * delta[1:, ...]

    # pressure at highest half level <= 0.1
    if np.any(p_half_level[0, ...] <= PRESSURE_TOA):
        alpha[0, ...] = 1.0  # ARPEGE choice, ECMWF IFS uses log(2)
    # pressure at highest half level > 0.1
    else:
        alpha[0, ...] = 1.0 - p_half_level[0, ...] / (p_half_level[1, ...] - p_half_level[0, ...]) * delta[0, ...]

    # calculate pressure on model full levels
    # TODO: is there a faster way to calculate the averages?
    # TODO: introduce option to calculate full levels in more complicated way
    p_full_level = np.apply_along_axis(lambda m: np.convolve(m, np.ones(2) / 2, mode="valid"), axis=0, arr=p_half_level)

    return p_full_level, p_half_level, delta, alpha


def calc_specific_gas_constant(q):
    """Calculates the specific gas constant of moist air
    (specific content of cloud particles and hydrometeors are neglected)

    Parameters
    ----------
    q : number or ndarray
        specific humidity

    Returns
    -------
    number or ndarray
        specific gas constant of moist air
    """

    R = constants.Rd + (constants.Rv - constants.Rd) * q
    return R


def relative_geopotential_thickness(alpha, q, T):
    """Calculates the geopotential thickness w.r.t the surface on model full-levels

    Parameters
    ----------
    alpha : ndarray
        alpha term of pressure calculations
    q : ndarray
        specific humidity (in kg/kg) on model full-levels
    T : ndarray
        temperature (in Kelvin) on model full-levels

    Returns
    -------
    ndarray
        geopotential thickness of model full-levels w.r.t. the surface
    """

    R = calc_specific_gas_constant(q)
    dphi = np.cumsum(np.flip(alpha * R * T, axis=0), axis=0)
    dphi = np.flip(dphi, axis=0)

    return dphi


def pressure_at_height_level(height, q, T, sp, A, B):
    """Calculates the pressure at a height level given in meters above surface.
    This is done by finding the model level above and below the specified height
    and interpolating the pressure

    Parameters
    ----------
    height : number
        height (in meters) above the surface for which the pressure is wanted
    q : ndarray
        specific humidity (kg/kg) at model full-levels
    T : ndarray
        temperature (K) at model full-levels
    sp : ndarray
        surface pressure (Pa)
    A : ndarray
        A-coefficients defining the model levels
    B : ndarray
        B-coefficients defining the model levels

    Returns
    -------
    number or ndarray
        pressure (Pa) at the given height level
    """

    # geopotential thickness of the height level
    tdphi = height * constants.g

    # pressure(-related) variables
    p_full, p_half, _, alpha = model_level_pressure(A, B, sp)

    # relative geopot. thickness of full levels
    dphi = relative_geopotential_thickness(alpha, q, T)

    # find the model full level right above the height level
    i_phi = (tdphi > dphi).sum(0)

    # initialize the output array
    p_height = np.zeros_like(i_phi, dtype=np.float64)

    # define mask: requested height is below the lowest model full-level
    mask = i_phi == 0

    # CASE 1: requested height is below the lowest model full-level
    # --> interpolation between surface pressure and lowest model full-level
    p_height[mask] = (p_half[-1, ...] + tdphi / dphi[-1, ...] * (p_full[-1, ...] - p_half[-1, ...]))[mask]

    # CASE 2: requested height is above the lowest model full-level
    # --> interpolation between between model full-level above and below

    # define some indices for masking and readability
    i_lev = alpha.shape[0] - i_phi - 1  # convert phi index to model level index
    indices = np.indices(i_lev.shape)
    masked_indices = tuple(dim[~mask] for dim in indices)
    above = (i_lev[~mask],) + masked_indices
    below = (i_lev[~mask] + 1,) + masked_indices

    dphi_above = dphi[above]
    dphi_below = dphi[below]

    factor = (tdphi - dphi_above) / (dphi_below - dphi_above)
    p_height[~mask] = p_full[above] + factor * (p_full[below] - p_full[above])

    return p_height


def execute(context, input, height, t, q, sp, new_name="2r", **kwargs):
    """Convert the single (height) level specific humidity to relative humidity"""
    result = FieldArray()

    MANDATORY_KEYS = ["A", "B"]
    OPTIONAL_KEYS = ["t_ml", "q_ml"]
    MISSING_KEYS = []
    DEFAULTS = dict(t_ml="t", q_ml="q")

    for key in OPTIONAL_KEYS:
        if key not in kwargs:
            print(f"key {key} not found in yaml-file, using default key: {DEFAULTS[key]}")
            kwargs[key] = DEFAULTS[key]

    for key in MANDATORY_KEYS:
        if key not in kwargs:
            MISSING_KEYS.append(key)

    if MISSING_KEYS:
        raise KeyError(f"Following keys are missing: {', '.join(MISSING_KEYS)}")

    single_level_params = (t, q, sp)
    model_level_params = (kwargs["t_ml"], kwargs["q_ml"])

    needed_fields = AutoDict()

    # Gather all necessary fields
    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        # check single level parameters
        if param in single_level_params:
            levtype = key.pop("levtype")
            key = tuple(sorted(key.items()))

            if param in needed_fields[key][levtype]:
                raise ValueError(f"Duplicate single level field {param} for {key}")

            needed_fields[key][levtype][param] = f
            if param == q:
                if kwargs.get("keep_q", False):
                    result.append(f)
            else:
                result.append(f)

        # check model level parameters
        elif param in model_level_params:
            levtype = key.pop("levtype")
            levelist = key.pop("levelist")
            key = tuple(sorted(key.items()))

            if param in needed_fields[key][levtype][levelist]:
                raise ValueError(f"Duplicate model level field {param} for {key} at level {levelist}")

            needed_fields[key][levtype][levelist][param] = f

        # all other parameters
        else:
            result.append(f)

    for _, values in needed_fields.items():
        # some checks
        if len(values["sfc"]) != 3:
            raise ValueError("Missing surface fields")

        q_sl = values["sfc"][q].to_numpy(flatten=True)
        t_sl = values["sfc"][t].to_numpy(flatten=True)
        sp_sl = values["sfc"][sp].to_numpy(flatten=True)

        nlevels = len(kwargs["A"]) - 1
        if len(values["ml"]) != nlevels:
            raise ValueError("Missing model levels")

        for key in values["ml"].keys():
            if len(values["ml"][key]) != 2:
                raise ValueError(f"Missing field on level {key}")

        # create 3D arrays for upper air fields
        levels = list(values["ml"].keys())
        levels.sort()
        t_ml = []
        q_ml = []
        for level in levels:
            t_ml.append(values["ml"][level][kwargs["t_ml"]].to_numpy(flatten=True))
            q_ml.append(values["ml"][level][kwargs["q_ml"]].to_numpy(flatten=True))

        t_ml = np.stack(t_ml)
        q_ml = np.stack(q_ml)

        # actual conversion from qv --> rh
        # FIXME:
        # For now We need to go from qv --> td --> rh to take into account
        # the mixed / ice phase when T ~ 0C / T < 0C
        # See https://github.com/ecmwf/earthkit-meteo/issues/15
        p_sl = pressure_at_height_level(height, q_ml, t_ml, sp_sl, np.array(kwargs["A"]), np.array(kwargs["B"]))
        td_sl = thermo.dewpoint_from_specific_humidity(q=q_sl, p=p_sl)
        rh_sl = thermo.relative_humidity_from_dewpoint(t=t_sl, td=td_sl)

        result.append(NewDataField(values["sfc"][q], rh_sl, new_name))

    return result


def test():
    from earthkit.data import from_source
    from earthkit.data.readers.grib.index import GribFieldList

    # IFS forecasts have both specific humidity and dewpoint
    sl = from_source(
        "mars",
        {
            "date": "2022-01-01",
            "class": "od",
            "expver": "1",
            "stream": "oper",
            "levtype": "sfc",
            "param": "96.174/134.128/167.128/168.128",
            "time": "00:00:00",
            "type": "fc",
            "step": "2",
            "grid": "O640",
        },
    )

    ml = from_source(
        "mars",
        {
            "date": "2022-01-01",
            "class": "od",
            "expver": "1",
            "stream": "oper",
            "levtype": "ml",
            "levelist": "130/131/132/133/134/135/136/137",
            "param": "130/133",
            "time": "00:00:00",
            "type": "fc",
            "step": "2",
            "grid": "O640",
        },
    )
    source = GribFieldList.merge([sl, ml])

    # IFS A and B coeffients for level 137 - 129
    kwargs = {
        "A": [424.414063, 302.476563, 202.484375, 122.101563, 62.781250, 22.835938, 3.757813, 0.0, 0.0],
        "B": [0.969513, 0.975078, 0.980072, 0.984542, 0.988500, 0.991984, 0.995003, 0.997630, 1.000000],
    }
    source = execute(None, source, 2, "2t", "2sh", "sp", "2r", **kwargs)

    temperature = source[2].to_numpy(flatten=True)
    dewpoint = source[3].to_numpy(flatten=True)
    relhum = source[4].to_numpy()
    newdew = thermo.dewpoint_from_relative_humidity(temperature, relhum)

    print(f"Mean difference in dewpoint temperature: {np.abs(newdew - dewpoint).mean():02f} degC")
    print(f"Median difference in dewpoint temperature: {np.median(np.abs(newdew - dewpoint)):02f} degC")
    print(f"Maximum difference in dewpoint temperature: {np.abs(newdew - dewpoint).max():02f} degC")

    # source.save("source.grib")


if __name__ == "__main__":
    test()



from anemoi.datasets import open_dataset

config = {
    "dataset": {
        "ensemble": [
            "/path/to/dataset1.zarr",
            {"dataset": "dataset_name", "end": 2010},
            {"dataset": "s3://path/to/dataset3.zarr", "start": 2000, "end": 2010},
        ],
        "frequency": "24h",
    },
    "select": ["2t", "msl"],
}

ds = open_dataset(config)



from anemoi.datasets import add_dataset_path
from anemoi.datasets import open_dataset

add_dataset_path("https://object-store.os-api.cci1.ecmwf.int/ml-examples/")

ds = open_dataset("an-oper-2023-2023-2p5-6h-v1")



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict

import numpy as np
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.meteo.wind.array import xy_to_polar

from anemoi.datasets.create.functions.filters.speeddir_to_uv import NewDataField


def execute(context, input, u_component, v_component, wind_speed, wind_dir, in_radians=False):
    result = FieldArray()

    wind_params = (u_component, v_component)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    for _, pairs in wind_pairs.items():
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        u = pairs[u_component]
        v = pairs[v_component]

        # assert speed.grid_mapping == dir.grid_mapping
        magnitude, direction = xy_to_polar(u.to_numpy(flatten=True), v.to_numpy(flatten=True))
        if in_radians:
            direction = np.deg2rad(direction)

        result.append(NewDataField(u, magnitude, wind_speed))
        result.append(NewDataField(v, direction, wind_dir))

    return result



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from copy import deepcopy
from functools import cached_property

from anemoi.datasets.dates import DatesProvider

from .action import Action
from .action import action_factory
from .empty import EmptyResult
from .misc import _tidy
from .misc import assert_fieldlist
from .result import Result
from .template import notify_result
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class ConcatResult(Result):
    def __init__(self, context, action_path, group_of_dates, results, **kwargs):
        super().__init__(context, action_path, group_of_dates)
        self.results = [r for r in results if not r.empty]

    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self):
        ds = EmptyResult(self.context, self.action_path, self.group_of_dates).datasource
        for i in self.results:
            ds += i.datasource
        return _tidy(ds)

    @property
    def variables(self):
        """Check that all the results objects have the same variables."""
        variables = None
        for f in self.results:
            if f.empty:
                continue
            if variables is None:
                variables = f.variables
            assert variables == f.variables, (variables, f.variables)
        assert variables is not None, self.results
        return variables

    def __repr__(self):
        content = "\n".join([str(i) for i in self.results])
        return super().__repr__(content)


class ConcatAction(Action):
    def __init__(self, context, action_path, *configs):
        super().__init__(context, action_path, *configs)
        parts = []
        for i, cfg in enumerate(configs):
            if "dates" not in cfg:
                raise ValueError(f"Missing 'dates' in {cfg}")
            cfg = deepcopy(cfg)
            dates_cfg = cfg.pop("dates")
            assert isinstance(dates_cfg, dict), dates_cfg
            filtering_dates = DatesProvider.from_config(**dates_cfg)
            action = action_factory(cfg, context, action_path + [str(i)])
            parts.append((filtering_dates, action))
        self.parts = parts

    def __repr__(self):
        content = "\n".join([str(i) for i in self.parts])
        return super().__repr__(content)

    @trace_select
    def select(self, group_of_dates):
        from anemoi.datasets.dates.groups import GroupOfDates

        results = []
        for filtering_dates, action in self.parts:
            newdates = GroupOfDates(sorted(set(group_of_dates) & set(filtering_dates)), group_of_dates.provider)
            if newdates:
                results.append(action.select(newdates))
        if not results:
            return EmptyResult(self.context, self.action_path, group_of_dates)

        return ConcatResult(self.context, self.action_path, group_of_dates, results)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from earthkit.data.core.fieldlist import MultiFieldList
from earthkit.data.sources.url import download_and_cache

from . import iterate_patterns
from .xarray import load_one


def execute(context, dates, record_id, file_key, *args, **kwargs):
    import requests

    result = []

    URLPATTERN = "https://zenodo.org/api/records/{record_id}"
    url = URLPATTERN.format(record_id=record_id)
    r = requests.get(url)
    r.raise_for_status()
    record = r.json()

    urls = {}
    for file in record["files"]:
        urls[file["key"]] = file["links"]["self"]

    for url, dates in iterate_patterns(file_key, dates, **kwargs):
        if url not in urls:
            continue

        path = download_and_cache(urls[url])
        result.append(load_one("?", context, dates, path, options={}, flavour=None, **kwargs))

    return MultiFieldList(result)



#!/usr/bin/env python
# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import tqdm
import zarr

from anemoi.datasets import open_dataset

from . import Command


class Compare(Command):
    """Compare two datasets. This command compares the variables in two datasets and prints the mean of the common variables. It does not compare the data itself (yet)."""

    def add_arguments(self, command_parser):
        command_parser.add_argument("dataset1")
        command_parser.add_argument("dataset2")
        command_parser.add_argument("--data", action="store_true", help="Compare the data.")
        command_parser.add_argument("--statistics", action="store_true", help="Compare the statistics.")

    def run(self, args):
        ds1 = open_dataset(args.dataset1)
        ds2 = open_dataset(args.dataset2)

        v1 = set(ds1.variables)
        v2 = set(ds2.variables)

        print("Only in dataset 1:", ", ".join(sorted(v1 - v2)))
        print("Only in dataset 2:", ", ".join(sorted(v2 - v1)))
        print()
        common = sorted(v1 & v2)
        print("Common:")
        print("-------")
        print()

        for v in common:
            print(
                f"{v:14}",
                f"{ds1.statistics['mean'][ds1.name_to_index[v]]:14g}",
                f"{ds2.statistics['mean'][ds2.name_to_index[v]]:14g}",
            )

        if args.data:
            print()
            print("Data:")
            print("-----")
            print()

            diff = 0
            for a, b in tqdm.tqdm(zip(ds1, ds2)):
                if not np.array_equal(a, b, equal_nan=True):
                    diff += 1

            print(f"Number of different rows: {diff}/{len(ds1)}")

        if args.data:
            print()
            print("Data 2:")
            print("-----")
            print()

            ds1 = zarr.open(args.dataset1, mode="r")
            ds2 = zarr.open(args.dataset2, mode="r")

            for name in (
                "data",
                "count",
                "sums",
                "squares",
                "mean",
                "stdev",
                "minimum",
                "maximum",
                "latitudes",
                "longitudes",
            ):
                a1 = ds1[name]
                a2 = ds2[name]

                if len(a1) != len(a2):
                    print(f"{name}: lengths mismatch {len(a1)} != {len(a2)}")
                    continue

                diff = 0
                for a, b in tqdm.tqdm(zip(a1, a2), leave=False):
                    if not np.array_equal(a, b, equal_nan=True):
                        if diff == 0:
                            print(f"\n{name}: first different row:")
                            print(a[a != b])
                            print(b[a != b])

                        diff += 1

                print(f"{name}: {diff} different rows out of {len(a1)}")


command = Compare



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import textwrap

from anemoi.utils.dates import as_datetime as as_datetime
from anemoi.utils.dates import frequency_to_timedelta as frequency_to_timedelta
from anemoi.utils.humanize import plural

from anemoi.datasets.dates import DatesProvider as DatesProvider
from anemoi.datasets.fields import FieldArray as FieldArray
from anemoi.datasets.fields import NewValidDateTimeField as NewValidDateTimeField

from .trace import step
from .trace import trace

LOG = logging.getLogger(__name__)


class Context:
    def __init__(self):
        # used_references is a set of reference paths that will be needed
        self.used_references = set()
        # results is a dictionary of reference path -> obj
        self.results = {}

    def will_need_reference(self, key):
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        self.used_references.add(key)

    def notify_result(self, key, result):
        trace(
            "🎯",
            step(key),
            "notify result",
            textwrap.shorten(repr(result).replace(",", ", "), width=40),
            plural(len(result), "field"),
        )
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.used_references:
            if key in self.results:
                raise ValueError(f"Duplicate result {key}")
            self.results[key] = result

    def get_result(self, key):
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.results:
            return self.results[key]
        all_keys = sorted(list(self.results.keys()))
        raise ValueError(f"Cannot find result {key} in {all_keys}")



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import requests
import xarray as xr

from anemoi.datasets.create.functions.sources.xarray import XarrayFieldList
from anemoi.datasets.testing import assert_field_list

URL = "https://object-store.os-api.cci1.ecmwf.int/ml-tests/test-data/samples/"

SAMPLES = list(range(23))
SKIP = [0, 1, 2, 3, 4]


def _test_samples(n, check_skip=True):

    r = requests.get(f"{URL}sample-{n:04d}.json")
    if r.status_code not in [200, 404]:
        r.raise_for_status()

    if r.status_code == 404:
        kwargs = {}
    else:
        kwargs = r.json()

    skip = kwargs.pop("skip", None)
    if skip and check_skip:
        pytest.skip(skip if isinstance(skip, str) else "Skipping test")

    open_zarr_kwargs = kwargs.pop("open_zarr_kwargs", {})

    ds = xr.open_zarr(f"{URL}sample-{n:04d}.zarr", consolidated=True, **open_zarr_kwargs)

    print(ds)

    from_xarray_kwargs = kwargs.pop("from_xarray_kwargs", {})

    fs = XarrayFieldList.from_xarray(ds, **from_xarray_kwargs)

    assert_field_list(fs, **kwargs)


@pytest.mark.parametrize("n", SAMPLES)
def test_samples(n):
    _test_samples(n)


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        for n in sys.argv[1:]:
            _test_samples(int(n), check_skip=False)
    else:
        for s in SAMPLES:
            if s in SKIP:
                continue
            print("-------------------", s)
            _test_samples(s, check_skip=False)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from functools import cached_property

from anemoi.utils.dates import as_datetime as as_datetime
from anemoi.utils.dates import frequency_to_timedelta as frequency_to_timedelta

from anemoi.datasets.dates import DatesProvider as DatesProvider
from anemoi.datasets.fields import FieldArray as FieldArray
from anemoi.datasets.fields import NewValidDateTimeField as NewValidDateTimeField

from .action import Action
from .action import action_factory
from .misc import _tidy
from .result import Result

LOG = logging.getLogger(__name__)


class DataSourcesAction(Action):
    def __init__(self, context, action_path, sources, input):
        super().__init__(context, ["data_sources"], *sources)
        if isinstance(sources, dict):
            configs = [(str(k), c) for k, c in sources.items()]
        elif isinstance(sources, list):
            configs = [(str(i), c) for i, c in enumerate(sources)]
        else:
            raise ValueError(f"Invalid data_sources, expecting list or dict, got {type(sources)}: {sources}")

        self.sources = [action_factory(config, context, ["data_sources"] + [a_path]) for a_path, config in configs]
        self.input = action_factory(input, context, ["input"])

    def select(self, group_of_dates):
        sources_results = [a.select(group_of_dates) for a in self.sources]
        return DataSourcesResult(
            self.context,
            self.action_path,
            group_of_dates,
            self.input.select(group_of_dates),
            sources_results,
        )

    def __repr__(self):
        content = "\n".join([str(i) for i in self.sources])
        return super().__repr__(content)


class DataSourcesResult(Result):
    def __init__(self, context, action_path, dates, input_result, sources_results):
        super().__init__(context, action_path, dates)
        # result is the main input result
        self.input_result = input_result
        # sources_results is the list of the sources_results
        self.sources_results = sources_results

    @cached_property
    def datasource(self):
        for i in self.sources_results:
            # for each result trigger the datasource to be computed
            # and saved in context
            self.context.notify_result(i.action_path[:-1], i.datasource)
        # then return the input result
        # which can use the datasources of the included results
        return _tidy(self.input_result.datasource)



import logging

from . import Command

LOG = logging.getLogger(__name__)


class Publish(Command):
    """Publish a dataset."""

    # This is a command that is used to publish a dataset.
    # it is a class, inheriting from Command.

    internal = True
    timestamp = True

    def add_arguments(self, parser):
        parser.add_argument("path", help="Path of the dataset to publish.")

    def run(self, args):
        try:
            from anemoi.registry import publish_dataset
        except ImportError:
            LOG.error("anemoi-registry is not installed. Please install it to use this command.")
            return

        publish_dataset(args.path)


command = Publish



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

LOG = logging.getLogger(__name__)


class ViewCacheArray:
    """A class that provides a caching mechanism for writing to a NumPy-like array.

    The is initialized with a NumPy-like array, a shape and a list to reindex the first
    dimension. The array is used to store the final data, while the cache is used to
    temporarily store the data before flushing it to the array.

    The `flush` method copies the contents of the cache to the final array.

    """

    def __init__(self, array, *, shape, indexes):
        assert len(indexes) == shape[0], (len(indexes), shape[0])
        self.array = array
        self.dtype = array.dtype
        self.cache = np.full(shape, np.nan, dtype=self.dtype)
        self.indexes = indexes

    def __setitem__(self, key, value):
        self.cache[key] = value

    def flush(self):
        for i in range(self.cache.shape[0]):
            global_i = self.indexes[i]
            self.array[global_i] = self.cache[i]



#!/usr/bin/env python3


import argparse
import os
import shutil

import xarray as xr

parser = argparse.ArgumentParser(description="Create a sample dataset")
parser.add_argument("input", type=str, help="Input file name")
parser.add_argument("output", type=str, help="Output file name")
args = parser.parse_args()

if os.path.exists(args.output):
    if os.path.isdir(args.output):
        shutil.rmtree(args.output)
    else:
        os.unlink(args.output)

if args.input.endswith(".zarr"):
    ds = xr.open_zarr(args.input)
else:
    ds = xr.open_dataset(args.input)

if args.output.endswith(".zarr"):
    ds.to_zarr(args.output, consolidated=True)
else:
    ds.to_netcdf(args.output)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from earthkit.data.indexing.fieldlist import FieldArray


def list_to_fieldlist(fields):
    return FieldArray(fields)


def empty_fieldlist():
    return FieldArray([])


class WrappedField:
    def __init__(self, field):
        self._field = field

    def __getattr__(self, name):
        return getattr(self._field, name)

    def __repr__(self) -> str:
        return repr(self._field)


class NewDataField(WrappedField):
    def __init__(self, field, data):
        super().__init__(field)
        self._data = data
        self.shape = data.shape

    def to_numpy(self, flatten=False, dtype=None, index=None):
        data = self._data
        if dtype is not None:
            data = data.astype(dtype)
        if flatten:
            data = data.flatten()
        if index is not None:
            data = data[index]
        return data


class NewMetadataField(WrappedField):
    def __init__(self, field, **kwargs):
        super().__init__(field)
        self._metadata = kwargs

    def metadata(self, *args, **kwargs):
        if len(args) == 1 and args[0] in self._metadata:
            return self._metadata[args[0]]
        return self._field.metadata(*args, **kwargs)


class NewValidDateTimeField(NewMetadataField):
    def __init__(self, field, valid_datetime):
        date = valid_datetime.date().strftime("%Y%m%d")
        time = valid_datetime.time().strftime("%H%M")

        self.valid_datetime = valid_datetime

        super().__init__(field, date=date, time=time, step=0, valid_datetime=valid_datetime.isoformat())



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re
from functools import wraps

LOG = logging.getLogger(__name__)


def notify_result(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        self.context.notify_result(self.action_path, result)
        return result

    return wrapper


class Substitution:
    pass


class Reference(Substitution):
    def __init__(self, context, action_path):
        self.context = context
        self.action_path = action_path

    def resolve(self, context):
        return context.get_result(self.action_path)


def resolve(context, x):
    if isinstance(x, tuple):
        return tuple([resolve(context, y) for y in x])

    if isinstance(x, list):
        return [resolve(context, y) for y in x]

    if isinstance(x, dict):
        return {k: resolve(context, v) for k, v in x.items()}

    if isinstance(x, Substitution):
        return x.resolve(context)

    return x


def substitute(context, x):
    if isinstance(x, tuple):
        return tuple([substitute(context, y) for y in x])

    if isinstance(x, list):
        return [substitute(context, y) for y in x]

    if isinstance(x, dict):
        return {k: substitute(context, v) for k, v in x.items()}

    if not isinstance(x, str):
        return x

    if re.match(r"^\${[\.\w]+}$", x):
        path = x[2:-1].split(".")
        context.will_need_reference(path)
        return Reference(context, path)

    return x



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import glob
import logging

from earthkit.data.utils.patterns import Pattern

LOG = logging.getLogger(__name__)


def _expand(paths):

    if not isinstance(paths, list):
        paths = [paths]

    for path in paths:
        if path.startswith("file://"):
            path = path[7:]

        if path.startswith("http://"):
            yield path
            continue

        if path.startswith("https://"):
            yield path
            continue

        cnt = 0
        for p in glob.glob(path):
            yield p
            cnt += 1
        if cnt == 0:
            yield path


def iterate_patterns(path, dates, **kwargs):
    given_paths = path if isinstance(path, list) else [path]

    dates = [d.isoformat() for d in dates]
    if len(dates) > 0:
        kwargs["date"] = dates

    for path in given_paths:
        paths = Pattern(path, ignore_missing_keys=True).substitute(**kwargs)
        for path in _expand(paths):
            yield path, dates



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from .action import Action
from .action import action_factory
from .step import step_factory
from .trace import trace_select

LOG = logging.getLogger(__name__)


class PipeAction(Action):
    def __init__(self, context, action_path, *configs):
        super().__init__(context, action_path, *configs)
        assert len(configs) > 1, configs
        current = action_factory(configs[0], context, action_path + ["0"])
        for i, c in enumerate(configs[1:]):
            current = step_factory(c, context, action_path + [str(i + 1)], previous_step=current)
        self.last_step = current

    @trace_select
    def select(self, group_of_dates):
        return self.last_step.select(group_of_dates)

    def __repr__(self):
        return super().__repr__(self.last_step)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np
from anemoi.utils.dates import frequency_to_timedelta

from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class InterpolateFrequency(Forwards):

    def __init__(self, dataset, frequency):
        super().__init__(dataset)
        self._frequency = frequency_to_timedelta(frequency)

        self.seconds = self._frequency.total_seconds()
        other_seconds = dataset.frequency.total_seconds()

        self.seconds = int(self.seconds)
        assert self.seconds == self._frequency.total_seconds()

        other_seconds = int(other_seconds)
        assert other_seconds == dataset.frequency.total_seconds()

        if self.seconds >= other_seconds:
            raise ValueError(
                f"Interpolate frequency {self._frequency} must be more frequent than dataset frequency {dataset.frequency}"
            )

        if other_seconds % self.seconds != 0:
            raise ValueError(
                f"Interpolate frequency {self._frequency}  must be a multiple of the dataset frequency {dataset.frequency}"
            )

        self.ratio = other_seconds // self.seconds
        self.alphas = np.linspace(0, 1, self.ratio + 1)
        self.other_len = len(dataset)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 0, slice(None))
        result = self._get_slice(previous)
        return apply_index_to_slices_changes(result[index], changes)

    def _get_slice(self, s):
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        if n < 0:
            n += self._len

        if n == self._len - 1:
            # Special case for the last element
            return self.forward[-1]

        i = n // self.ratio
        x = n % self.ratio

        if x == 0:
            # No interpolation needed
            return self.forward[i]

        alpha = self.alphas[x]

        assert 0 < alpha < 1, alpha
        return self.forward[i] * (1 - alpha) + self.forward[i + 1] * alpha

    def __len__(self):
        return (self.other_len - 1) * self.ratio + 1

    @property
    def frequency(self):
        return self._frequency

    @cached_property
    def dates(self):
        result = []
        deltas = [np.timedelta64(self.seconds * i, "s") for i in range(self.ratio)]
        for d in self.forward.dates[:-1]:
            for i in deltas:
                result.append(d + i)
        result.append(self.forward.dates[-1])
        return np.array(result)

    @property
    def shape(self):
        return (self._len,) + self.forward.shape[1:]

    def tree(self):
        return Node(self, [self.forward.tree()], frequency=self.frequency)

    @cached_property
    def missing(self):
        result = []
        j = 0
        for i in range(self.other_len):
            missing = i in self.forward.missing
            for _ in range(self.ratio):
                if missing:
                    result.append(j)
                j += 1

        result = set(x for x in result if x < self._len)
        return result

    def subclass_metadata_specific(self):
        return {
            # "frequency": frequency_to_string(self._frequency),
        }



import datetime
import logging
import time
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

import tqdm
from anemoi.utils.humanize import seconds_to_human

from . import Command

LOG = logging.getLogger(__name__)


def task(what, options, *args, **kwargs):
    """
    Make sure `import Creator` is done in the sub-processes, and not in the main one.
    """

    now = datetime.datetime.now()
    LOG.info(f"🎬 Task {what}({args},{kwargs}) starting")

    from anemoi.datasets.create import creator_factory

    options = {k: v for k, v in options.items() if v is not None}

    c = creator_factory(what.replace("-", "_"), **options)
    result = c.run()

    LOG.info(f"🏁 Task {what}({args},{kwargs}) completed ({datetime.datetime.now()-now})")
    return result


class Create(Command):
    """Create a dataset."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing files. This will delete the target dataset if it already exists.",
        )
        command_parser.add_argument(
            "--test",
            action="store_true",
            help="Build a small dataset, using only the first dates. And, when possible, using low resolution and less ensemble members.",
        )
        command_parser.add_argument("config", help="Configuration yaml file defining the recipe to create the dataset.")
        command_parser.add_argument("path", help="Path to store the created data.")
        group = command_parser.add_mutually_exclusive_group()
        group.add_argument("--threads", help="Use `n` parallel thread workers.", type=int, default=0)
        group.add_argument("--processes", help="Use `n` parallel process workers.", type=int, default=0)
        command_parser.add_argument("--trace", action="store_true")

    def run(self, args):

        now = time.time()
        if args.threads + args.processes:
            self.parallel_create(args)
        else:
            self.serial_create(args)
        LOG.info(f"Create completed in {seconds_to_human(time.time()-now)}")

    def serial_create(self, args):

        options = vars(args)
        options.pop("command")
        options.pop("threads")
        options.pop("processes")

        task("init", options)
        task("load", options)
        task("finalise", options)

        task("patch", options)

        task("init_additions", options)
        task("run_additions", options)
        task("finalise_additions", options)
        task("cleanup", options)
        task("verify", options)

    def parallel_create(self, args):
        """Some modules, like fsspec do not work well with fork()
        Other modules may not be thread safe. So we implement
        parallel loadining using multiprocessing before any
        of the modules are imported.
        """

        options = vars(args)
        options.pop("command")

        threads = options.pop("threads")
        processes = options.pop("processes")

        use_threads = threads > 0
        options["use_threads"] = use_threads

        if use_threads:
            ExecutorClass = ThreadPoolExecutor
        else:
            ExecutorClass = ProcessPoolExecutor

        with ExecutorClass(max_workers=1) as executor:
            total = executor.submit(task, "init", options).result()

        futures = []

        parallel = threads + processes
        with ExecutorClass(max_workers=parallel) as executor:
            for n in range(total):
                opt = options.copy()
                opt["parts"] = f"{n+1}/{total}"
                futures.append(executor.submit(task, "load", opt))

            for future in tqdm.tqdm(as_completed(futures), desc="Loading", total=len(futures), colour="green", position=parallel + 1):  # fmt: skip
                future.result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "finalise", options).result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "init-additions", options).result()

        with ExecutorClass(max_workers=parallel) as executor:
            opt = options.copy()
            opt["parts"] = f"{n+1}/{total}"
            futures.append(executor.submit(task, "load", opt))
            for n in range(total):
                futures.append(executor.submit(task, "load-additions", opt))

            for future in tqdm.tqdm(as_completed(futures), desc="Computing additions", total=len(futures), colour="green", position=parallel + 1):  # fmt: skip
                future.result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "finalise-additions", options).result()
            executor.submit(task, "cleanup", options).result()
            executor.submit(task, "verify", options).result()


command = Create



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np

from anemoi.datasets.data.indexing import length_to_slices


def test_length_to_slices():
    lengths = [5, 7, 11, 13]
    datasets = [np.random.rand(n) for n in lengths]
    total = sum(lengths)

    combined = np.concatenate(datasets)

    for start in range(total):
        for stop in range(start, total):
            for step in range(1, stop - start + 1):
                index = slice(start, stop, step)
                print(index)
                slices = length_to_slices(index, lengths)
                result = [d[i] for (d, i) in zip(datasets, slices) if i is not None]
                result = np.concatenate(result)

                if (combined[index].shape != result.shape) or not (combined[index] == result).all():
                    print(index)
                    print(combined[index])
                    print(result)
                    print(slices)
                assert (combined[index] == result).all(), index


if __name__ == "__main__":
    test_length_to_slices()



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import textwrap
import threading
from functools import wraps

LOG = logging.getLogger(__name__)


thread_local = threading.local()
TRACE = 0


def enable_trace(on_off):
    global TRACE
    TRACE = on_off


def step(action_path):
    return f"[{'.'.join(action_path)}]"


def trace(emoji, *args):

    if not TRACE:
        return

    if not hasattr(thread_local, "TRACE_INDENT"):
        thread_local.TRACE_INDENT = 0

    print(emoji, " " * thread_local.TRACE_INDENT, *args)


def trace_datasource(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):

        if not hasattr(thread_local, "TRACE_INDENT"):
            thread_local.TRACE_INDENT = 0

        trace(
            "🌍",
            "=>",
            step(self.action_path),
            self._trace_datasource(*args, **kwargs),
        )
        thread_local.TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        thread_local.TRACE_INDENT -= 1
        trace(
            "🍎",
            "<=",
            step(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper


def trace_select(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not hasattr(thread_local, "TRACE_INDENT"):
            thread_local.TRACE_INDENT = 0
        trace(
            "👓",
            "=>",
            ".".join(self.action_path),
            self._trace_select(*args, **kwargs),
        )
        thread_local.TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        thread_local.TRACE_INDENT -= 1
        trace(
            "🍍",
            "<=",
            ".".join(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import datetime
import logging

from anemoi.utils.dates import as_datetime

LOG = logging.getLogger(__name__)


class Time:

    @classmethod
    def from_coordinates(cls, coordinates):
        time_coordinate = [c for c in coordinates if c.is_time]
        step_coordinate = [c for c in coordinates if c.is_step]
        date_coordinate = [c for c in coordinates if c.is_date]

        if len(date_coordinate) == 0 and len(time_coordinate) == 1 and len(step_coordinate) == 1:
            return ForecastFromValidTimeAndStep(time_coordinate[0], step_coordinate[0])

        if len(date_coordinate) == 0 and len(time_coordinate) == 1 and len(step_coordinate) == 0:
            return Analysis(time_coordinate[0])

        if len(date_coordinate) == 0 and len(time_coordinate) == 0 and len(step_coordinate) == 0:
            return Constant()

        if len(date_coordinate) == 1 and len(time_coordinate) == 1 and len(step_coordinate) == 0:
            return ForecastFromValidTimeAndBaseTime(date_coordinate[0], time_coordinate[0])

        if len(date_coordinate) == 1 and len(time_coordinate) == 0 and len(step_coordinate) == 1:
            return ForecastFromBaseTimeAndDate(date_coordinate[0], step_coordinate[0])

        if len(date_coordinate) == 1 and len(time_coordinate) == 1 and len(step_coordinate) == 1:
            return ForecastFromValidTimeAndStep(time_coordinate[0], step_coordinate[0], date_coordinate[0])

        LOG.error("")
        LOG.error(f"{len(date_coordinate)} date_coordinate")
        for c in date_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        LOG.error("")
        LOG.error(f"{len(time_coordinate)} time_coordinate")
        for c in time_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        LOG.error("")
        LOG.error(f"{len(step_coordinate)} step_coordinate")
        for c in step_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        raise NotImplementedError(f"{len(date_coordinate)=} {len(time_coordinate)=} {len(step_coordinate)=}")


class Constant(Time):

    def fill_time_metadata(self, coords_values, metadata):
        return None


class Analysis(Time):

    def __init__(self, time_coordinate):
        self.time_coordinate_name = time_coordinate.variable.name

    def fill_time_metadata(self, coords_values, metadata):
        valid_datetime = coords_values[self.time_coordinate_name]

        metadata["date"] = as_datetime(valid_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(valid_datetime).strftime("%H%M")
        metadata["step"] = 0

        return valid_datetime


class ForecastFromValidTimeAndStep(Time):

    def __init__(self, time_coordinate, step_coordinate, date_coordinate=None):
        self.time_coordinate_name = time_coordinate.variable.name
        self.step_coordinate_name = step_coordinate.variable.name
        self.date_coordinate_name = date_coordinate.variable.name if date_coordinate else None

    def fill_time_metadata(self, coords_values, metadata):
        valid_datetime = coords_values[self.time_coordinate_name]
        step = coords_values[self.step_coordinate_name]

        assert isinstance(step, datetime.timedelta)
        base_datetime = valid_datetime - step

        hours = step.total_seconds() / 3600
        assert int(hours) == hours

        metadata["date"] = as_datetime(base_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(base_datetime).strftime("%H%M")
        metadata["step"] = int(hours)

        # When date is present, it should be compatible with time and step

        if self.date_coordinate_name is not None:
            # Not sure that this is the correct assumption
            assert coords_values[self.date_coordinate_name] == base_datetime, (
                coords_values[self.date_coordinate_name],
                base_datetime,
            )

        return valid_datetime


class ForecastFromValidTimeAndBaseTime(Time):

    def __init__(self, date_coordinate, time_coordinate):
        self.date_coordinate.name = date_coordinate.name
        self.time_coordinate.name = time_coordinate.name

    def fill_time_metadata(self, coords_values, metadata):
        valid_datetime = coords_values[self.time_coordinate_name]
        base_datetime = coords_values[self.date_coordinate_name]

        step = valid_datetime - base_datetime

        hours = step.total_seconds() / 3600
        assert int(hours) == hours

        metadata["date"] = as_datetime(base_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(base_datetime).strftime("%H%M")
        metadata["step"] = int(hours)

        return valid_datetime


class ForecastFromBaseTimeAndDate(Time):

    def __init__(self, date_coordinate, step_coordinate):
        self.date_coordinate_name = date_coordinate.name
        self.step_coordinate_name = step_coordinate.name

    def fill_time_metadata(self, coords_values, metadata):

        date = coords_values[self.date_coordinate_name]
        step = coords_values[self.step_coordinate_name]
        assert isinstance(step, datetime.timedelta)

        metadata["date"] = as_datetime(date).strftime("%Y%m%d")
        metadata["time"] = as_datetime(date).strftime("%H%M")

        hours = step.total_seconds() / 3600

        assert int(hours) == hours
        metadata["step"] = int(hours)

        return date + step



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from earthkit.data.core.fieldlist import MultiFieldList

from anemoi.datasets.create.functions.sources.mars import mars

LOGGER = logging.getLogger(__name__)


def _to_list(x):
    if isinstance(x, (list, tuple)):
        return x
    return [x]


def hindcasts(context, dates, **request):

    from anemoi.datasets.dates import HindcastsDates

    provider = context.dates_provider
    assert isinstance(provider, HindcastsDates)

    context.trace("H️", f"hindcasts {len(dates)=}")

    request["param"] = _to_list(request["param"])
    request["step"] = _to_list(request.get("step", 0))
    request["step"] = [int(_) for _ in request["step"]]

    context.trace("H️", f"hindcast {request}")

    requests = []
    for d in dates:
        r = request.copy()
        hindcast = provider.mapping[d]
        r["hdate"] = hindcast.hdate.strftime("%Y-%m-%d")
        r["date"] = hindcast.refdate.strftime("%Y-%m-%d")
        r["time"] = hindcast.refdate.strftime("%H")
        r["step"] = hindcast.step
        requests.append(r)

    if len(requests) == 0:
        return MultiFieldList([])

    return mars(
        context,
        dates,
        *requests,
        date_key="hdate",
        request_already_using_valid_datetime=True,
    )


execute = hindcasts



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

import tqdm
from anemoi.utils.humanize import bytes_to_human

LOG = logging.getLogger(__name__)


def compute_directory_sizes(path):
    if not os.path.isdir(path):
        return None

    size, n = 0, 0
    bar = tqdm.tqdm(iterable=os.walk(path), desc=f"Computing size of {path}")
    for dirpath, _, filenames in bar:
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            size += os.path.getsize(file_path)
            n += 1

    LOG.info(f"Total size: {bytes_to_human(size)}")
    LOG.info(f"Total number of files: {n}")

    return dict(total_size=size, total_number_of_files=n)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import itertools
import logging
import math
import time
from collections import defaultdict
from copy import deepcopy
from functools import cached_property
from functools import wraps

import numpy as np
from anemoi.utils.dates import as_datetime as as_datetime
from anemoi.utils.dates import frequency_to_timedelta as frequency_to_timedelta

from anemoi.datasets.dates import DatesProvider as DatesProvider
from anemoi.datasets.fields import FieldArray as FieldArray
from anemoi.datasets.fields import NewValidDateTimeField as NewValidDateTimeField

from .trace import trace_select

LOG = logging.getLogger(__name__)


class InputBuilder:
    def __init__(self, config, data_sources, **kwargs):
        self.kwargs = kwargs

        config = deepcopy(config)
        if data_sources:
            config = dict(
                data_sources=dict(
                    sources=data_sources,
                    input=config,
                )
            )
        self.config = config
        self.action_path = ["input"]

    @trace_select
    def select(self, group_of_dates):
        from .action import ActionContext
        from .action import action_factory

        """This changes the context."""
        context = ActionContext(**self.kwargs)
        action = action_factory(self.config, context, self.action_path)
        return action.select(group_of_dates)

    def __repr__(self):
        from .action import ActionContext
        from .action import action_factory

        context = ActionContext(**self.kwargs)
        a = action_factory(self.config, context, self.action_path)
        return repr(a)

    def _trace_select(self, group_of_dates):
        return f"InputBuilder({group_of_dates})"


build_input = InputBuilder



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
from functools import cached_property

import numpy as np

LOG = logging.getLogger(__name__)


class Grid:

    def __init__(self):
        pass

    @property
    def latitudes(self):
        return self.grid_points[0]

    @property
    def longitudes(self):
        return self.grid_points[1]


class LatLonGrid(Grid):
    def __init__(self, lat, lon, variable_dims):
        super().__init__()
        self.lat = lat
        self.lon = lon


class XYGrid(Grid):
    def __init__(self, x, y):
        self.x = x
        self.y = y


class MeshedGrid(LatLonGrid):

    @cached_property
    def grid_points(self):

        lat, lon = np.meshgrid(
            self.lat.variable.values,
            self.lon.variable.values,
        )

        return lat.flatten(), lon.flatten()


class UnstructuredGrid(LatLonGrid):

    def __init__(self, lat, lon, variable_dims):
        super().__init__(lat, lon, variable_dims)
        assert len(lat) == len(lon), (len(lat), len(lon))
        self.variable_dims = variable_dims
        self.grid_dims = lat.variable.dims
        assert lon.variable.dims == self.grid_dims, (lon.variable.dims, self.grid_dims)
        assert set(self.variable_dims) == set(self.grid_dims), (self.variable_dims, self.grid_dims)

    @cached_property
    def grid_points(self):

        assert 1 <= len(self.variable_dims) <= 2

        if len(self.variable_dims) == 1:
            return self.lat.variable.values.flatten(), self.lon.variable.values.flatten()

        if len(self.variable_dims) == 2 and self.variable_dims == self.grid_dims:
            return self.lat.variable.values.flatten(), self.lon.variable.values.flatten()

        LOG.warning(
            "UnstructuredGrid: variable indexing %s does not match grid indexing %s", self.variable_dims, self.grid_dims
        )

        lat = self.lat.variable.values.transpose().flatten()
        lon = self.lon.variable.values.transpose().flatten()

        return lat, lon


class ProjectionGrid(XYGrid):
    def __init__(self, x, y, projection):
        super().__init__(x, y)
        self.projection = projection

    def transformer(self):
        from pyproj import CRS
        from pyproj import Transformer

        if isinstance(self.projection, dict):
            data_crs = CRS.from_cf(self.projection)
        else:
            data_crs = self.projection

        wgs84_crs = CRS.from_epsg(4326)  # WGS84

        return Transformer.from_crs(data_crs, wgs84_crs, always_xy=True)


class MeshProjectionGrid(ProjectionGrid):

    @cached_property
    def grid_points(self):

        transformer = self.transformer()
        xv, yv = np.meshgrid(self.x.variable.values, self.y.variable.values)  # , indexing="ij")
        lon, lat = transformer.transform(xv, yv)
        return lat.flatten(), lon.flatten()


class UnstructuredProjectionGrid(XYGrid):
    @cached_property
    def grid_points(self):
        assert False, "Not implemented"

        # lat, lon = transformer.transform(
        #      self.y.variable.values.flatten(),
        #     self.x.variable.values.flatten(),

        # )

        # lat = lat[::len(lat)//100]
        # lon = lon[::len(lon)//100]

        # print(len(lat), len(lon))

        # return np.meshgrid(lat, lon)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from functools import cached_property

from anemoi.utils.dates import as_datetime as as_datetime
from anemoi.utils.dates import frequency_to_timedelta as frequency_to_timedelta

from anemoi.datasets.dates import DatesProvider as DatesProvider
from anemoi.datasets.fields import FieldArray as FieldArray
from anemoi.datasets.fields import NewValidDateTimeField as NewValidDateTimeField

from ..functions import import_function
from .action import Action
from .misc import _tidy
from .misc import assert_fieldlist
from .result import Result
from .template import notify_result
from .template import resolve
from .template import substitute
from .trace import trace
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class FunctionContext:
    """A FunctionContext is passed to all functions, it will be used to pass information
    to the functions from the other actions and filters and results.
    """

    def __init__(self, owner):
        self.owner = owner
        self.use_grib_paramid = owner.context.use_grib_paramid

    def trace(self, emoji, *args):
        trace(emoji, *args)

    def info(self, *args, **kwargs):
        LOG.info(*args, **kwargs)

    @property
    def dates_provider(self):
        return self.owner.group_of_dates.provider

    @property
    def partial_ok(self):
        return self.owner.group_of_dates.partial_ok


class FunctionAction(Action):
    def __init__(self, context, action_path, _name, **kwargs):
        super().__init__(context, action_path, **kwargs)
        self.name = _name

    @trace_select
    def select(self, group_of_dates):
        return FunctionResult(self.context, self.action_path, group_of_dates, action=self)

    @property
    def function(self):
        # name, delta = parse_function_name(self.name)
        return import_function(self.name, "sources")

    def __repr__(self):
        content = ""
        content += ",".join([self._short_str(a) for a in self.args])
        content += " ".join([self._short_str(f"{k}={v}") for k, v in self.kwargs.items()])
        content = self._short_str(content)
        return super().__repr__(_inline_=content, _indent_=" ")

    def _trace_select(self, group_of_dates):
        return f"{self.name}({group_of_dates})"


class FunctionResult(Result):
    def __init__(self, context, action_path, group_of_dates, action):
        super().__init__(context, action_path, group_of_dates)
        assert isinstance(action, Action), type(action)
        self.action = action

        self.args, self.kwargs = substitute(context, (self.action.args, self.action.kwargs))

    def _trace_datasource(self, *args, **kwargs):
        return f"{self.action.name}({self.group_of_dates})"

    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self):
        args, kwargs = resolve(self.context, (self.args, self.kwargs))

        try:
            return _tidy(
                self.action.function(
                    FunctionContext(self),
                    list(self.group_of_dates),  # Will provide a list of datetime objects
                    *args,
                    **kwargs,
                )
            )
        except Exception:
            LOG.error(f"Error in {self.action.function.__name__}", exc_info=True)
            raise

    def __repr__(self):
        try:
            return f"{self.action.name}({self.group_of_dates})"
        except Exception:
            return f"{self.__class__.__name__}(unitialised)"

    @property
    def function(self):
        raise NotImplementedError(f"Not implemented in {self.__class__.__name__}")



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
import os

import xarray as xr
from multiurl import download

from anemoi.datasets.create.functions.sources.xarray import XarrayFieldList

URLS = {
    "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/efas.nc": dict(length=3),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/era5-2m-temperature-dec-1993.nc": dict(length=1),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.nc": dict(length=2),
    "https://get.ecmwf.int/repository/test-data/metview/gallery/era5_2000_aug.nc": dict(length=3),
    "https://get.ecmwf.int/repository/test-data/metview/gallery/era5_ozone_1999.nc": dict(length=4),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/fa_ta850.nc": dict(length=37),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/htessel_points.nc": dict(length=1),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/test_single.nc": dict(length=1),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/zgrid_rhgmet_metop_200701_R_2305_0010.nc": dict(
        length=1
    ),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/20210101-C3S-L2_GHG-GHG_PRODUCTS-TANSO2-GOSAT2-SRFP-DAILY-v2.0.0.nc": dict(
        length=1
    ),
    "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/20220401-C3S-L3S_FIRE-BA-OLCI-AREA_3-fv1.1.nc": dict(
        length=3
    ),
    "https://github.com/ecmwf/magics-test/raw/master/test/efas/tamir.nc": dict(length=1),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/C3S_OZONE-L4-TC-ASSIM_MSR-201608-fv0020.nc": dict(
        length=1
    ),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/avg_data.nc": dict(length=1),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/era5_2000_aug_1.nc": dict(length=3),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/missing.nc": dict(length=20),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/netcdf3_t_z.nc": dict(length=30),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/tos_O1_2001-2002.nc": dict(length=24),
    "https://github.com/ecmwf/magics-test/raw/master/test/gallery/z_500.nc": dict(length=1),
}


def skip_test_netcdf():

    for url, checks in URLS.items():
        print(url)
        path = os.path.join(os.path.dirname(__file__), os.path.basename(url))
        if not os.path.exists(path):
            download(url, path)

        ds = xr.open_dataset(path)

        fs = XarrayFieldList.from_xarray(ds)

        assert len(fs) == checks["length"], (url, len(fs))

        print(fs[0].datetime())


if __name__ == "__main__":
    skip_test_netcdf()



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import warnings

# from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import DateTimes
from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta
from anemoi.utils.hindcasts import HindcastDatesTimes
from anemoi.utils.humanize import print_dates


def extend(x):

    if isinstance(x, (list, tuple)):
        for y in x:
            yield from extend(y)
        return

    if isinstance(x, str):
        if "/" in x:
            start, end, step = x.split("/")
            start = as_datetime(start)
            end = as_datetime(end)
            step = frequency_to_timedelta(step)
            while start <= end:
                yield start
                start += step
            return

    yield as_datetime(x)


class DatesProvider:
    """Base class for date generation.

    >>> DatesProvider.from_config(**{"start": "2023-01-01 00:00", "end": "2023-01-02 00:00", "frequency": "1d"}).values
    [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 2, 0, 0)]

    >>> DatesProvider.from_config(**{"start": "2023-01-01 00:00", "end": "2023-01-03 00:00", "frequency": "18h"}).values
    [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 1, 18, 0), datetime.datetime(2023, 1, 2, 12, 0)]

    >>> DatesProvider.from_config(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=6).as_dict()
    {'start': '2023-01-01T00:00:00', 'end': '2023-01-02T00:00:00', 'frequency': '6h'}

    >>> len(DatesProvider.from_config(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=12))
    3
    >>> len(DatesProvider.from_config(start="2023-01-01 00:00",
    ...                   end="2023-01-02 00:00",
    ...                   frequency=12,
    ...                   missing=["2023-01-01 12:00"]))
    3
    >>> len(DatesProvider.from_config(start="2023-01-01 00:00",
    ...                   end="2023-01-02 00:00",
    ...                   frequency=12,
    ...                   missing=["2099-01-01 12:00"]))
    3
    """

    def __init__(self, missing=None):
        if not missing:
            missing = []
        self.missing = list(extend(missing))
        if set(self.missing) - set(self.values):
            diff = set(self.missing) - set(self.values)
            warnings.warn(f"Missing dates {len(diff)=} not in list.")

    @classmethod
    def from_config(cls, **kwargs):

        if kwargs.pop("hindcasts", False):
            return HindcastsDates(**kwargs)

        if "values" in kwargs:
            return ValuesDates(**kwargs)

        return StartEndDates(**kwargs)

    def __iter__(self):
        yield from self.values

    def __getitem__(self, i):
        return self.values[i]

    def __len__(self):
        return len(self.values)

    @property
    def summary(self):
        return f"📅 {self.values[0]} ... {self.values[-1]}"


class ValuesDates(DatesProvider):
    def __init__(self, values, **kwargs):
        self.values = sorted([as_datetime(_) for _ in values])
        super().__init__(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values[0]}..{self.values[-1]})"

    def as_dict(self):
        return {"values": self.values[0]}


class StartEndDates(DatesProvider):
    def __init__(self, start, end, frequency=1, **kwargs):

        frequency = frequency_to_timedelta(frequency)
        assert isinstance(frequency, datetime.timedelta), frequency

        def _(x):
            if isinstance(x, str):
                return datetime.datetime.fromisoformat(x)
            return x

        start = _(start)
        end = _(end)

        if isinstance(start, datetime.date) and not isinstance(start, datetime.datetime):
            start = datetime.datetime(start.year, start.month, start.day)

        if isinstance(end, datetime.date) and not isinstance(end, datetime.datetime):
            end = datetime.datetime(end.year, end.month, end.day)

        start = as_datetime(start)
        end = as_datetime(end)

        self.start = start
        self.end = end
        self.frequency = frequency

        missing = kwargs.pop("missing", [])

        self.values = list(DateTimes(start, end, increment=frequency, **kwargs))
        self.kwargs = kwargs

        super().__init__(missing=missing)

    def as_dict(self):
        return {
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "frequency": frequency_to_string(self.frequency),
        }.update(self.kwargs)


class Hindcast:

    def __init__(self, date, refdate, hdate, step):
        self.date = date
        self.refdate = refdate
        self.hdate = hdate
        self.step = step


class HindcastsDates(DatesProvider):
    def __init__(self, start, end, steps=[0], years=20, **kwargs):

        if not isinstance(start, list):
            start = [start]
            end = [end]

        reference_dates = []
        for s, e in zip(start, end):
            reference_dates.extend(list(DateTimes(s, e, increment=24, **kwargs)))
        # reference_dates = list(DateTimes(start, end, increment=24, **kwargs))
        dates = []

        seen = {}

        for hdate, refdate in HindcastDatesTimes(reference_dates=reference_dates, years=years):
            assert refdate - hdate >= datetime.timedelta(days=365), (refdate - hdate, refdate, hdate)
            for step in steps:

                date = hdate + datetime.timedelta(hours=step)

                if date in seen:
                    raise ValueError(f"Duplicate date {date}={hdate}+{step} for {refdate} and {seen[date]}")

                seen[date] = Hindcast(date, refdate, hdate, step)

                assert refdate - date > datetime.timedelta(days=360), (refdate - date, refdate, date, hdate, step)

                dates.append(date)

        dates = sorted(dates)

        mindelta = None
        for a, b in zip(dates, dates[1:]):
            delta = b - a
            assert isinstance(delta, datetime.timedelta), delta
            if mindelta is None:
                mindelta = delta
            else:
                mindelta = min(mindelta, delta)

        self.frequency = mindelta
        assert mindelta.total_seconds() > 0, mindelta

        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", dates[0], dates[-1], mindelta)

        # Use all values between start and end by frequency, and set the ones that are missing
        self.values = []
        missing = []
        date = dates[0]
        last = date
        print("------", date, dates[-1])
        dateset = set(dates)
        while date <= dates[-1]:
            self.values.append(date)
            if date not in dateset:
                missing.append(date)
                seen[date] = seen[last]
            else:
                last = date
            date = date + mindelta

        self.mapping = seen

        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", self.values[0], self.values[-1], mindelta)
        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", f"{len(self.values)=} - {len(missing)=}")

        super().__init__(missing=missing)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values[0]}..{self.values[-1]})"

    def as_dict(self):
        return {"hindcasts": self.hindcasts}


if __name__ == "__main__":
    print_dates([datetime.datetime(2023, 1, 1, 0, 0)])
    s = StartEndDates(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=1)
    print_dates(list(s))



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from functools import wraps

from anemoi.utils.dates import as_datetime as as_datetime
from anemoi.utils.dates import frequency_to_timedelta as frequency_to_timedelta
from earthkit.data.core.fieldlist import MultiFieldList
from earthkit.data.indexing.fieldlist import FieldList

from anemoi.datasets.dates import DatesProvider as DatesProvider
from anemoi.datasets.fields import FieldArray as FieldArray
from anemoi.datasets.fields import NewValidDateTimeField as NewValidDateTimeField

from ..functions import import_function

LOG = logging.getLogger(__name__)


def parse_function_name(name):

    if name.endswith("h") and name[:-1].isdigit():

        if "-" in name:
            name, delta = name.split("-")
            sign = -1

        elif "+" in name:
            name, delta = name.split("+")
            sign = 1

        else:
            return name, None

        assert delta[-1] == "h", (name, delta)
        delta = sign * int(delta[:-1])
        return name, delta

    return name, None


def is_function(name, kind):
    name, _ = parse_function_name(name)
    try:
        import_function(name, kind)
        return True
    except ImportError as e:
        print(e)
        return False


def assert_fieldlist(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        assert isinstance(result, FieldList), type(result)
        return result

    return wrapper


def assert_is_fieldlist(obj):
    assert isinstance(obj, FieldList), type(obj)


def _flatten(ds):
    if isinstance(ds, MultiFieldList):
        return [_tidy(f) for s in ds._indexes for f in _flatten(s)]
    return [ds]


def _tidy(ds, indent=0):
    if isinstance(ds, MultiFieldList):

        sources = [s for s in _flatten(ds) if len(s) > 0]
        if len(sources) == 1:
            return sources[0]
        return MultiFieldList(sources)
    return ds



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from earthkit.data import from_source


def constants(context, dates, template, param):
    from warnings import warn

    warn(
        "The source `constants` is deprecated, use `forcings` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    context.trace("✅", f"from_source(constants, {template}, {param}")
    if len(template) == 0:
        raise ValueError("Forcings template is empty.")

    return from_source("forcings", source_or_dataset=template, date=dates, param=param)


execute = constants



import logging
import time

from anemoi.utils.humanize import seconds_to_human

from anemoi.datasets.commands.create import task

from . import Command

LOG = logging.getLogger(__name__)


class Init(Command):
    """Create a dataset, step by step."""

    internal = True
    timestamp = True

    def add_arguments(self, subparser):

        subparser.add_argument("config", help="Configuration yaml file defining the recipe to create the dataset.")
        subparser.add_argument("path", help="Path to store the created data.")

        subparser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing files. This will delete the target dataset if it already exists.",
        )
        subparser.add_argument(
            "--test",
            action="store_true",
            help="Build a small dataset, using only the first dates. And, when possible, using low resolution and less ensemble members.",
        )
        subparser.add_argument(
            "--check-name",
            dest="check_name",
            action="store_true",
            help="Check if the dataset name is valid before creating it.",
        )
        subparser.add_argument(
            "--no-check-name",
            dest="check_name",
            action="store_false",
            help="Do not check if the dataset name is valid before creating it.",
        )
        subparser.set_defaults(check_name=False)
        subparser.add_argument("--cache", help="Location to store the downloaded data.", metavar="DIR")

        subparser.add_argument("--trace", action="store_true")

    def run(self, args):
        options = vars(args)
        options.pop("command")
        now = time.time()

        if "version" in options:
            options.pop("version")

        if "debug" in options:
            options.pop("debug")

        task("init", options)

        LOG.info(f"Create step '{self.__class__.__name__.lower()}' completed in {seconds_to_human(time.time()-now)}")


command = Init



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output

from anemoi.datasets.create.functions import assert_is_fieldlist

LOG = logging.getLogger(__name__)

CLIP_VARIABLES = (
    "q",
    "cp",
    "lsp",
    "tp",
    "sf",
    "swl4",
    "swl3",
    "swl2",
    "swl1",
)

SKIP = ("class", "stream", "type", "number", "expver", "_leg_number", "anoffset", "time", "date", "step")


def check_compatible(f1, f2, centre_field_as_mars, ensemble_field_as_mars):
    assert f1.mars_grid == f2.mars_grid, (f1.mars_grid, f2.mars_grid)
    assert f1.mars_area == f2.mars_area, (f1.mars_area, f2.mars_area)
    assert f1.shape == f2.shape, (f1.shape, f2.shape)

    # Not in *_as_mars
    assert f1.metadata("valid_datetime") == f2.metadata("valid_datetime"), (
        f1.metadata("valid_datetime"),
        f2.metadata("valid_datetime"),
    )

    for k in set(centre_field_as_mars.keys()) | set(ensemble_field_as_mars.keys()):
        if k in SKIP:
            continue
        assert centre_field_as_mars[k] == ensemble_field_as_mars[k], (
            k,
            centre_field_as_mars[k],
            ensemble_field_as_mars[k],
        )


def recentre(
    *,
    members,
    centre,
    clip_variables=CLIP_VARIABLES,
    alpha=1.0,
    output=None,
):

    keys = ["param", "level", "valid_datetime", "date", "time", "step", "number"]

    number_list = members.unique_values("number", progress_bar=False)["number"]
    n_numbers = len(number_list)

    assert None not in number_list

    LOG.info("Ordering fields")
    members = members.order_by(*keys)
    centre = centre.order_by(*keys)
    LOG.info("Done")

    if len(centre) * n_numbers != len(members):
        LOG.error("%s %s %s", len(centre), n_numbers, len(members))
        for f in members:
            LOG.error("Member: %r", f)
        for f in centre:
            LOG.error("centre: %r", f)
        raise ValueError(f"Inconsistent number of fields: {len(centre)} * {n_numbers} != {len(members)}")

    if output is None:
        # prepare output tmp file so we can read it back
        tmp = temp_file()
        path = tmp.path
    else:
        tmp = None
        path = output

    out = new_grib_output(path)

    seen = set()

    for i, centre_field in enumerate(centre):
        param = centre_field.metadata("param")
        centre_field_as_mars = centre_field.metadata(namespace="mars")

        # load the centre field
        centre_np = centre_field.to_numpy()

        # load the ensemble fields and compute the mean
        members_np = np.zeros((n_numbers, *centre_np.shape))

        for j in range(n_numbers):
            ensemble_field = members[i * n_numbers + j]
            ensemble_field_as_mars = ensemble_field.metadata(namespace="mars")
            check_compatible(
                centre_field,
                ensemble_field,
                centre_field_as_mars,
                ensemble_field_as_mars,
            )
            members_np[j] = ensemble_field.to_numpy()

            ensemble_field_as_mars = tuple(sorted(ensemble_field_as_mars.items()))
            assert ensemble_field_as_mars not in seen, ensemble_field_as_mars
            seen.add(ensemble_field_as_mars)

        # cmin=np.amin(centre_np)
        # emin=np.amin(members_np)

        # if cmin < 0 and emin >= 0:
        #     LOG.warning(f"Negative values in {param} cmin={cmin} emin={emin}")
        #     LOG.warning(f"centre: {centre_field_as_mars}")

        mean_np = members_np.mean(axis=0)

        for j in range(n_numbers):
            template = members[i * n_numbers + j]
            e = members_np[j]
            m = mean_np
            c = centre_np

            assert e.shape == c.shape == m.shape, (e.shape, c.shape, m.shape)

            x = c + (e - m) * alpha

            if param in clip_variables:
                # LOG.warning(f"Clipping {param} to be positive")
                x = np.maximum(x, 0)

            assert x.shape == e.shape, (x.shape, e.shape)

            out.write(x, template=template)
            template = None

    assert len(seen) == len(members), (len(seen), len(members))

    out.close()

    if output is not None:
        return path

    from earthkit.data import from_source

    ds = from_source("file", path)
    assert_is_fieldlist(ds)
    # save a reference to the tmp file so it is deleted
    # only when the dataset is not used anymore
    ds._tmp = tmp

    assert len(ds) == len(members), (len(ds), len(members))

    return ds



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import warnings
from functools import cached_property
from urllib.parse import urlparse

import numpy as np
import zarr
from anemoi.utils.dates import frequency_to_timedelta

from . import MissingDateError
from .dataset import Dataset
from .debug import DEBUG_ZARR_LOADING
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .indexing import expand_list_indexing
from .misc import load_config

LOG = logging.getLogger(__name__)


class ReadOnlyStore(zarr.storage.BaseStore):
    def __delitem__(self, key):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()


class HTTPStore(ReadOnlyStore):
    """We write our own HTTPStore because the one used by zarr (s3fs)
    does not play well with fork() and multiprocessing.
    """

    def __init__(self, url):
        self.url = url

    def __getitem__(self, key):
        import requests

        r = requests.get(self.url + "/" + key)

        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return r.content


class S3Store(ReadOnlyStore):
    """We write our own S3Store because the one used by zarr (s3fs)
    does not play well with fork(). We also get to control the s3 client
    options using the anemoi configs.
    """

    def __init__(self, url, region=None):
        from anemoi.utils.s3 import s3_client

        _, _, self.bucket, self.key = url.split("/", 3)
        self.s3 = s3_client(self.bucket, region=region)

    def __getitem__(self, key):
        try:
            response = self.s3.get_object(Bucket=self.bucket, Key=self.key + "/" + key)
        except self.s3.exceptions.NoSuchKey:
            raise KeyError(key)

        return response["Body"].read()


class DebugStore(ReadOnlyStore):
    """A store to debug the zarr loading."""

    def __init__(self, store):
        assert not isinstance(store, DebugStore)
        self.store = store

    def __getitem__(self, key):
        # print()
        print("GET", key, self)
        # traceback.print_stack(file=sys.stdout)
        return self.store[key]

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        warnings.warn("DebugStore: iterating over the store")
        return iter(self.store)

    def __contains__(self, key):
        return key in self.store


def name_to_zarr_store(path_or_url):
    store = path_or_url

    if store.startswith("s3://"):
        store = S3Store(store)

    elif store.startswith("http://") or store.startswith("https://"):
        parsed = urlparse(store)
        bits = parsed.netloc.split(".")
        if len(bits) == 5 and (bits[1], bits[3], bits[4]) == ("s3", "amazonaws", "com"):
            s3_url = f"s3://{bits[0]}{parsed.path}"
            store = S3Store(s3_url, region=bits[2])
        else:
            store = HTTPStore(store)

    return store


def open_zarr(path, dont_fail=False, cache=None):
    try:
        store = name_to_zarr_store(path)

        if DEBUG_ZARR_LOADING:
            if isinstance(store, str):
                import os

                if not os.path.isdir(store):
                    raise NotImplementedError(
                        "DEBUG_ZARR_LOADING is only implemented for DirectoryStore. "
                        "Please disable it for other backends."
                    )
                store = zarr.storage.DirectoryStore(store)
            store = DebugStore(store)

        if cache is not None:
            store = zarr.LRUStoreCache(store, max_size=cache)

        return zarr.convenience.open(store, "r")
    except zarr.errors.PathNotFoundError:
        if not dont_fail:
            raise zarr.errors.PathNotFoundError(path)


class Zarr(Dataset):
    """A zarr dataset."""

    def __init__(self, path):
        if isinstance(path, zarr.hierarchy.Group):
            self.was_zarr = True
            self.path = str(id(path))
            self.z = path
        else:
            self.was_zarr = False
            self.path = str(path)
            self.z = open_zarr(self.path)

        # This seems to speed up the reading of the data a lot
        self.data = self.z.data
        self.missing = set()

    @classmethod
    def from_name(cls, name):
        if name.endswith(".zip") or name.endswith(".zarr"):
            return Zarr(name)
        return Zarr(zarr_lookup(name))

    def __len__(self):
        return self.data.shape[0]

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n):
        return self.data[n]

    def _unwind(self, index, rest, shape, axis, axes):
        if not isinstance(index, (int, slice, list, tuple)):
            try:
                # NumPy arrays, TensorFlow tensors, etc.
                index = tuple(index.tolist())
                assert not isinstance(index, bool), "Mask not supported"
            except AttributeError:
                pass

        if isinstance(index, (list, tuple)):
            axes.append(axis)  # Dimension of the concatenation
            for i in index:
                yield from self._unwind((slice(i, i + 1),), rest, shape, axis, axes)
            return

        if len(rest) == 0:
            yield (index,)
            return

        for n in self._unwind(rest[0], rest[1:], shape, axis + 1, axes):
            yield (index,) + n

    @cached_property
    def chunks(self):
        return self.z.data.chunks

    @cached_property
    def shape(self):
        return self.data.shape

    @cached_property
    def dtype(self):
        return self.z.data.dtype

    @cached_property
    def dates(self):
        return self.z.dates[:]  # Convert to numpy

    @property
    def latitudes(self):
        try:
            return self.z.latitudes[:]
        except AttributeError:
            LOG.warning("No 'latitudes' in %r, trying 'latitude'", self)
            return self.z.latitude[:]

    @property
    def longitudes(self):
        try:
            return self.z.longitudes[:]
        except AttributeError:
            LOG.warning("No 'longitudes' in %r, trying 'longitude'", self)
            return self.z.longitude[:]

    @property
    def statistics(self):
        return dict(
            mean=self.z.mean[:],
            stdev=self.z.stdev[:],
            maximum=self.z.maximum[:],
            minimum=self.z.minimum[:],
        )

    def statistics_tendencies(self, delta=None):
        if delta is None:
            delta = self.frequency
        if isinstance(delta, int):
            delta = f"{delta}h"
        from anemoi.utils.dates import frequency_to_string
        from anemoi.utils.dates import frequency_to_timedelta

        delta = frequency_to_timedelta(delta)
        delta = frequency_to_string(delta)

        def func(k):
            return f"statistics_tendencies_{delta}_{k}"

        return dict(
            mean=self.z[func("mean")][:],
            stdev=self.z[func("stdev")][:],
            maximum=self.z[func("maximum")][:],
            minimum=self.z[func("minimum")][:],
        )

    @property
    def resolution(self):
        return self.z.attrs["resolution"]

    @property
    def field_shape(self):
        try:
            return tuple(self.z.attrs["field_shape"])
        except KeyError:
            LOG.warning("No 'field_shape' in %r, assuming 1D fields", self)
            return (self.shape[-1],)

    @property
    def frequency(self):
        try:
            return frequency_to_timedelta(self.z.attrs["frequency"])
        except KeyError:
            LOG.warning("No 'frequency' in %r, computing from 'dates'", self)
        dates = self.dates
        return dates[1].astype(object) - dates[0].astype(object)

    @property
    def name_to_index(self):
        if "variables" in self.z.attrs:
            return {n: i for i, n in enumerate(self.z.attrs["variables"])}
        return self.z.attrs["name_to_index"]

    @property
    def variables(self):
        return [
            k
            for k, v in sorted(
                self.name_to_index.items(),
                key=lambda x: x[1],
            )
        ]

    @property
    def variables_metadata(self):
        return self.z.attrs.get("variables_metadata", {})

    def __repr__(self):
        return self.path

    def end_of_statistics_date(self):
        return self.dates[-1]

    def metadata_specific(self):
        return super().metadata_specific(
            attrs=dict(self.z.attrs),
            chunks=self.chunks,
            dtype=str(self.dtype),
        )

    def source(self, index):
        return Source(self, index, info=self.path)

    def mutate(self):
        if len(self.z.attrs.get("missing_dates", [])):
            LOG.warning(f"Dataset {self} has missing dates")
            return ZarrWithMissingDates(self.z if self.was_zarr else self.path)
        return self

    def tree(self):
        return Node(self, [], path=self.path)

    def get_dataset_names(self, names):
        name, _ = os.path.splitext(os.path.basename(self.path))
        names.add(name)


class ZarrWithMissingDates(Zarr):
    """A zarr dataset with missing dates."""

    def __init__(self, path):
        super().__init__(path)

        missing_dates = self.z.attrs.get("missing_dates", [])
        missing_dates = set([np.datetime64(x) for x in missing_dates])
        self.missing_to_dates = {i: d for i, d in enumerate(self.dates) if d in missing_dates}
        self.missing = set(self.missing_to_dates)

    def mutate(self):
        return self

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n):
        if isinstance(n, int):
            if n in self.missing:
                self._report_missing(n)
            return self.data[n]

        if isinstance(n, slice):
            common = set(range(*n.indices(len(self)))) & self.missing
            if common:
                self._report_missing(list(common)[0])
            return self.data[n]

        if isinstance(n, tuple):
            first = n[0]
            if isinstance(first, int):
                if first in self.missing:
                    self._report_missing(first)
                return self.data[n]

            if isinstance(first, slice):
                common = set(range(*first.indices(len(self)))) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

            if isinstance(first, (list, tuple)):
                common = set(first) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

        raise TypeError(f"Unsupported index {n} {type(n)}")

    def _report_missing(self, n):
        raise MissingDateError(f"Date {self.missing_to_dates[n]} is missing (index={n})")

    def tree(self):
        return Node(self, [], path=self.path, missing=sorted(self.missing))

    @property
    def label(self):
        return "zarr*"


def zarr_lookup(name, fail=True):

    if name.endswith(".zarr") or name.endswith(".zip"):
        return name

    config = load_config()["datasets"]

    if name in config["named"]:
        return config["named"][name]

    tried = []
    for location in config["path"]:
        if not location.endswith("/"):
            location += "/"
        full = location + name + ".zarr"
        tried.append(full)
        try:
            z = open_zarr(full, dont_fail=True)
            if z is not None:
                # Cache for next time
                config["named"][name] = full
                return full
        except zarr.errors.PathNotFoundError:
            pass

    if fail:
        raise ValueError(f"Cannot find a dataset that matched '{name}'. Tried: {tried}")

    return None
